;;; popup-menu.el - Menu interface module for GNUemacs/NEmacs
;;; Copyright (C) HASHIMOTO Ryoichi
;;;        Industrial Products Research Institute, AIST/MITI Japan
;;;        Higashi 1-1-4, Tsukuba-shi, IBARAKI-ken, Zip(305), Japan
;;;        E-mail: hashimoto@ipri.go.jp
;;;        Tel: 0298-54-6661 (Desk)
;;;        Fax: 0298-54-6608

;;; This module provides features to support popup-menu interface
;;; in GNUemacs/NEmacs.  

;;; This is not an officially recognized part of GNU emacs nor NEmacs.
;;; But this must be used, copied, distributed, or modified under the
;;; same principle which GNU emacs and NEmacs assert.  

;;; You may freely copy, distribute, or modify this module, provided
;;; that: 

;;; 1) this copyrights note and comments are included without any 
;;;   modification; (you may add your comments, that's no problem.)  
;;; 2) you keep books, when you modify thid module, on the version 
;;;    history listed at the bottom of this comment field. In such a
;;;    case, add your initials to the version number so that cousin
;;;    versions can be easily distinguished.

;;; BUGS: As this is the first release, there is no bug reported so
;;; far.  If you find any bugs, feel complaints, find your solutions,
;;; or made any modification on this module, report it to the author,
;;; please, although he is not responsible to any bugs or malfunctions
;;; of this module.  Even comments or encouraging messages will be
;;; welcome.  He may respond them by chance :-)

;;; Version history:
;;;    alpha-1 1989-Sep-12 by HASHIMOTO Ryoichi, IPRI/AIST/MITI, Japan
;;;       Menu features and prepare-here
;;;    alpha-2 1989-Oct-14 by HASHIMOTO Ryoichi, IPRI/AIST/MITI, Japan
;;;       switch-select
;;;    1.00RH 1989-Nov-10 by HASHIMOTO Ryoichi, IPRI/AIST/MITI, japan
;;;    1.10RH 1989-Dec-19 by HASHIMOTO Ryoichi, IPRI/AIST/MITI, japan
;;;       DEBUG: Not to forget local-key after calling 'here-function

(provide 'popup-menu)

(setq *menu-escape-sequence* "\e\e")
(setq *menu-forward-char* nil)
(if *menu-forward-char*
    ()
  (cond
   ((eq (lookup-key global-map "\eOC") 'forward-char)
    (setq *menu-forward-char* "\eOC")
    (setq *menu-backward-char* "\eOD")
    (setq *menu-previous-line* "\eOA")
    (setq *menu-next-line* "\eOB")
   )
   ((eq (lookup-key global-map "\e[C") 'forward-char)
    (setq *menu-forward-char* "\e[C")
    (setq *menu-backward-char* "\e[D")
    (setq *menu-previous-line* "\e[A")
    (setq *menu-next-line* "\e[B")
   )
  (t
    (setq *menu-forward-char* "\C-f")
    (setq *menu-backward-char* "\C-b")
    (setq *menu-previous-line* "\C-p")
    (setq *menu-next-line* "\C-n")
   )
  )
)

(defun menu-define-both-key (mapname key-seq function)
  "Make key-bindings to both lower and capital"
  (define-key mapname key-seq function)
  (define-key mapname (capitalize key-seq) function)
)

(defun menu-setp (menu-candidate)
  "Returns t if MENU-CANDIDATE is a menu-set, otherwise nil." 
  (and
    (vectorp menu-candidate)
    (= (length menu-candidate) 5)
    (eq (aref menu-candidate 0) 'menu-set)
    (stringp (aref menu-candidate 1))
    (listp (aref menu-candidate 2))
    (keymapp (aref menu-candidate 3))
  )
)

(defun menu-define-key-append-menu-str (menu-set key-seq menu-str)
  "Replace or append menu string"
  (let ((tmp-list (aref menu-set 2))(tmp-cell nil))
    (setq tmp-cell (assoc key-seq tmp-list))
    (cond
      ((null tmp-list)
       (aset menu-set 2 (list (cons key-seq menu-str)))
      )
      ((null tmp-cell)
       (nconc (aref menu-set 2) (list (cons key-seq menu-str)))
      )
      (t (setcdr tmp-cell menu-str)
      )
    )
  )
)

(defun menu-define-key (menu-set key-seq func menu-str)
 "On MENU-SET, defines KEY-SEQ to FUNC.\
If KEY-SEQ is a single alpha-numeric, MENU-STR is registered.\
In such a case alpha must be in lower case.  If you violate this rule, \
it brings an inconsistence between menu and action."
 (cond
    ((null (menu-setp menu-set))
     (beep)
     (message "arg 1 to menu-define-key is not a menu-set")
    )
    ((and
      (= (length key-seq) 1)
      (or (= (elt key-seq 0) ? )
	  (and (>= (elt key-seq 0) ?0)
	       (<= (elt key-seq 0) ?9)
	  )
      )
     )
     (define-key
       (aref menu-set 3) key-seq
       (`(lambda () (interactive) (menu-exit)
	   (call-interactively (quote (, func)))
	 )
       )
     )
     (menu-define-key-append-menu-str menu-set key-seq menu-str)
    )
    ((and
      (= (length key-seq) 1)
      (>= (elt key-seq 0) ?a)
      (<= (elt key-seq 0) ?z)
     )
     (menu-define-both-key
       (aref menu-set 3) key-seq
       (`(lambda () (interactive) (menu-exit)
	   (call-interactively (quote (, func)))
	 )
       )
     )
     (menu-define-key-append-menu-str
          menu-set (capitalize key-seq) menu-str
     )
    )
    (t
      (define-key
	(aref menu-set 3) key-seq (menu-enmenu func)
        (`(lambda () (interactive) (menu-exit) 
	    (call-interactively (quote (, func)))
	  )
	)
      )
    )
  )
)

(defvar menu-var-window-config nil
 "window configuration before menu is triggered")

(defvar menu-var-local-map-before-menu nil
 "local keymap before menu is triggered")

(defun menu-display-menu (menu-set)
  "Displays menu-set and accepts choice"
  (interactive)
  (if (null (menu-setp menu-set))
      (message "arg to menu-display-menu is not a menu-set")
    (setq menu-var-window-config (current-window-configuration))
    (setq menu-var-local-map-before-menu (current-local-map))
    (switch-to-buffer (aref menu-set 1))
    (use-local-map (aref menu-set 3))
    (let ((width 3) (rest-of-menu-str (aref menu-set 2)) (its-car nil) (column-max 1))
      (if (aref menu-set 4)
	  (setq width (+ 3 (aref menu-set 4)))
	(setq width 80)
      )
      (setq column-max (/ 80 width))
      (setq column column-max)
      (while rest-of-menu-str
	(setq its-car (car rest-of-menu-str))
	(setq rest-of-menu-str (cdr rest-of-menu-str))
	(insert (format "%s:%s " (car its-car) (cdr its-car)))
	(setq column (1- column))
	(if (/= column 0)
	    ()
	  (insert "\n")
	  (setq column column-max)
	)
      )
    )
    (goto-char (point-min))
  )
)

(defun menu-exit ()
  "Exits menu normally or abnormally"
  (interactive)
  (kill-buffer (buffer-name))
  (set-window-configuration menu-var-window-config)
  (menu-resume-local-map)
)

(defun menu-select-this ()
  "Selects the alternative where cursor is located"
  (interactive)
  (if (and
         (< (point) (1- (point-max)))
	 (string= (buffer-substring (1+ (point)) (+ (point) 2)) ":")
      )
      ()
    (search-backward ":")
    (backward-char 1)
  )
  (let ((item-char (buffer-substring (point) (+ (point) 1))))
    (call-interactively 
      (lookup-key (or (current-local-map) (current-global-map)) item-char)
    )
  )
)

(defun menu-goto-previous-choice ()
  "Moves cursor to the previous alternative"
  (interactive)
  (if (<= (point) (1+ (point-min)))
      (beginning-of-buffer)
      (search-backward ":")
      (backward-char 1)
  )
)

(defun menu-goto-next-choice ()
  "Moves cursor to the next alternative"
  (interactive)
  (if (and
       (< (point) (- (point-max) 2))
       (progn (forward-char 2) (search-forward ":" (point-max) t))
      )
      (backward-char 2)
  )
)

(defun make-menu-set (menu-name &optional item-length)
 "Makes a menu-set named as the string arg MENU-NAME. \
If optional arg ITEM-LENGTH is given, declares to display short menu."
 (let ((tmp-map (make-keymap))(menu-set nil))
   (suppress-keymap tmp-map t)
   (setq menu-set
       (vector 'menu-set menu-name nil tmp-map item-length)
   )
   (define-key (aref menu-set 3) *menu-forward-char* 'menu-goto-next-choice)
   (define-key (aref menu-set 3) *menu-backward-char* 'menu-goto-previous-choice)
   (define-key (aref menu-set 3) *menu-escape-sequence* 'menu-exit)
   (define-key (aref menu-set 3) "\C-m" 'menu-select-this)
   (eval menu-set)
 )
)

(defun menu-resume-local-map ()
  "Discard current map for menu and resume the map for the buffer"
  (use-local-map menu-var-local-map-before-menu)
  (setq menu-var-local-map-before-menu nil)
)

;;;
;;; Execution point specification utilities
;;;

(defvar menu-here-map nil "")
(if menu-here-map
    ()
  (setq menu-here-map (make-keymap))
  (suppress-keymap menu-here-map)
)

(defun menu-prepare-here (func &optional cancel msg-str)
  "Prepares for the commands executed with specifyng the position.\
After this command you can invoke FUNC by \C-m. Optioanl second arg CANCEL
is undoing function. It can be nil.  Optional 3rd arg is prompt message string"
  (if (null cancel) (setq cancel 'undefined))
  (setq menu-var-local-map-before-menu (current-local-map))
  (use-local-map menu-here-map)
  (let ((local-var (list (cons "\C-m" func) (cons *menu-escape-sequence* cancel))))
    (while local-var
      (local-set-key (car (car local-var))
		 (`(lambda () (interactive)
		     (menu-resume-local-map)
		     (call-interactively (quote (, (cdr (car local-var)))))
		   )
		 )
      )
      (setq local-var (cdr local-var))
    )
  )
  (if (stringp msg-str) (message msg-str))
)
;;; menu switch list

;Example
;(defvar switch-list nil "")
;
;(setq jxw-switch-list
;      (cons "*Option List*"
;	    (list
;	     (cons 'jxw-beginning-prompt "prompt beginning of range: [%c]yes\t[%c]no")
;	     (cons 'jxw-execution-prompt "prompt execution point: [%c]yes\t[%c]no")
;	    )
;      )
;)

(defvar menu-switch-number 0)
(defvar menu-switch-list nil)
(defvar menu-switch-select-map nil "Which switch list to be selected")

(defun menu-switch-select (switch-list)
  "Display and change option switches.\n\
SWITCH-LIST is a dotted pair of <list name> and <list-body>,\n\
where <list body> is a list of <switch cell>'s.\n\
A <switch cell> is a dotted pair of <switch variable> and\n\ 
<displaying string> in the following format;\n\
\n\
  '<explanation>: [%c]<alternative1>\\t[%c]<alternative2>'.\n\
\n\
An example of displaying string> may be\n\
   'I like coffee: [%c]Yes\\t[%c]No'"
  (if menu-switch-select-map
      ()
    (setq menu-switch-select-map (make-keymap))
    (suppress-keymap menu-switch-select-map)
    (define-key menu-switch-select-map *menu-forward-char* 'menu-switch-negate)
    (define-key menu-switch-select-map *menu-backward-char* 'menu-switch-assert)
    (define-key menu-switch-select-map *menu-previous-line* 'menu-switch-previous)
    (define-key menu-switch-select-map *menu-next-line* 'menu-switch-next)
    (define-key menu-switch-select-map "\C-m" 'menu-switch-exit)
    (define-key menu-switch-select-map *menu-escape-sequence* 'menu-switch-exit)
  )
  (setq menu-var-window-config (current-window-configuration))
  (switch-to-buffer (car switch-list))
  (setq switch-list(cdr switch-list))
  (setq menu-switch-list switch-list)
  (while switch-list
    (menu-switch-rewrite-line (car switch-list) t)
    (setq switch-list (cdr switch-list))
  )
  (use-local-map menu-switch-select-map)
  (setq menu-switch-number -1)
  (beginning-of-buffer)
  (menu-switch-next)
)

(defun menu-switch-rewrite-line (item &optional cr)
  "Displays switch menu. Currently sellected alternative is shown by [O]."
  (if (= (point) (point-max))
      ()
    (beginning-of-line)
    (kill-line)
  )
  (if (eval (car item))
      (insert (format (cdr item) ?O ? ))
    (insert (format (cdr item) ?  ?O))
  )
  (if cr
      (insert "\n")
    (search-backward "\t")
  )
)

(defun menu-switch-negate ()
  "Set selected switch to be NIL"
  (interactive)
  (let ((switch-cell (nth menu-switch-number menu-switch-list)))
    (set (car switch-cell) nil)
    (menu-switch-rewrite-line switch-cell)
  )
)

(defun menu-switch-assert ()
  "Set selected switch to be t"
  (interactive)
  (let ((switch-cell (nth menu-switch-number menu-switch-list)))
    (set (car switch-cell) t)
    (menu-switch-rewrite-line switch-cell)
  )
)

(defun menu-switch-next ()
  "Select next item on the switch list"
  (interactive)
  (skip-chars-forward "\t")
  (if (search-forward "\t" (point-max) t)
      (setq menu-switch-number (1+ menu-switch-number))
  )
  (skip-chars-backward "\t")
)

(defun menu-switch-previous ()
  "Select previous item on the switch list"
  (interactive)
  (if (search-backward "\t" (point-min) t)
      (setq menu-switch-number (1- menu-switch-number))
  )
)

(defun menu-switch-exit ()
  "Storing selected switches, exits switch buffer."
  (interactive)
  (kill-buffer (buffer-name))
  (set-window-configuration menu-var-window-config)
)
;;; EOF
